@@ -10,34 +10,41 @@
 #include <gst/interfaces/xoverlay.h>
 
 #include "private.h"
-#include "cinema.h"
+#include "SDL_cinema.h"
 #include "debug.h"
 
 #define DEBUG_DOMAIN "CINEMA"
-#define TIMER_INTERVAL 1000
+
+static gboolean basic_init_done = FALSE;
 
 static Window window;
 static char *media_uri = NULL;
 static GstElement *pipeline = NULL;
 
-static SDL_TimerID timer_id = 0;
+static guint bus_watch_id = 0;
 
-void CIN_DeInit()
+static void free_pipeline()
 {
-	TRACE("DeInit");
+	if (bus_watch_id) {
+		g_source_remove(bus_watch_id);
+		bus_watch_id = 0;
+	}
 
 	if (pipeline) {
 		gst_element_set_state(pipeline, GST_STATE_NULL);
 		gst_object_unref(pipeline);
 		pipeline = NULL;
 	}
-	if (timer_id) {
-		SDL_RemoveTimer(timer_id);
-		timer_id = 0;
-	}
+}
+
+void CIN_DeInit()
+{
+	TRACE("DeInit");
+
+	free_pipeline();
+
 	g_free(media_uri);
 	media_uri = NULL;
-	gst_deinit();
 }
 
 int CIN_Init()
@@ -45,14 +52,17 @@ int CIN_Init()
 	SDL_SysWMinfo wminfo;
 	SDL_VERSION(&wminfo.version);
 
+	if (!basic_init_done) {
+		gst_init(NULL, NULL);
+		basic_init_done = TRUE;
+	}
+
 	if ( SDL_GetWMInfo(&wminfo) ) {
 		window = wminfo.info.x11.window;
 	} else {
 		return SDL_FALSE;
 	}
 
-	gst_init(NULL, NULL);
-
 	TRACE("Init; window xid is 0x%lx", (unsigned long) window);
 
 	return SDL_TRUE;
@@ -72,67 +82,49 @@ int CIN_LoadCIN(const char *url)
 	return SDL_TRUE;
 }
 
-static Uint32 timer_cb(Uint32 interval, void *param)
+static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
 {
-	GstBus *bus;
-	GstMessage *msg;
-	const GstMessageType msg_mask = GST_MESSAGE_ERROR | GST_MESSAGE_EOS;
+	switch (GST_MESSAGE_TYPE(msg)) {
+		case GST_MESSAGE_ERROR: {
+			SDL_Event event;
+			GError *err;
+			gchar *debug;
 
-	if (!pipeline) {
-		timer_id = 0;
-		return 0;
-	}
+			gst_message_parse_error(msg, &err, &debug);
+			WARN("GStreamer Pipeline Error: %s", err->message);
 
-	bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
-	if (!bus) {
-		timer_id = 0;
-		return 0;
-	}
+			event.type = SDL_USEREVENT;
+			event.user.code = SDL_CINEMAEVENT | SDL_CINEMA_ERR;
+			event.user.data1 = NULL;
+			event.user.data2 = NULL;
+			SDL_PushEvent(&event);
 
-	while ((msg = gst_bus_pop_filtered(bus, msg_mask))) {
-		switch (GST_MESSAGE_TYPE(msg)) {
-			case GST_MESSAGE_ERROR: {
-				SDL_Event event;
-				GError *err;
-				gchar *debug;
+			g_error_free(err);
+			g_free(debug);
 
-				gst_message_parse_error(msg, &err, &debug);
-				WARN("GStreamer Pipeline Error: %s", err->message);
-
-				event.type = SDL_USEREVENT;
-				event.user.code = SDL_CINEMAEVENT | SDL_CINEMA_ERR;
-				event.user.data1 = NULL;
-				event.user.data2 = NULL;
-				SDL_PushEvent(&event);
-
-				g_error_free(err);
-				g_free(debug);
-
-				timer_id = 0;
-				return 0;
-			}
-			break;
-			case GST_MESSAGE_EOS: {
-				TRACE("GStreamer Pipeline EOS");
-
-				SDL_Event event;
-				event.type = SDL_USEREVENT;
-				event.user.code = SDL_CINEMAEVENT | SDL_CINEMA_EOS;
-				event.user.data1 = NULL;
-				event.user.data2 = NULL;
-				SDL_PushEvent(&event);
-
-				timer_id = 0;
-				return 0;
-			}
-			break;
-			default:
-			break;
+			bus_watch_id = 0;
+			return FALSE;
 		}
-		gst_message_unref(msg);
+		break;
+		case GST_MESSAGE_EOS: {
+			TRACE("GStreamer Pipeline EOS");
+
+			SDL_Event event;
+			event.type = SDL_USEREVENT;
+			event.user.code = SDL_CINEMAEVENT | SDL_CINEMA_EOS;
+			event.user.data1 = NULL;
+			event.user.data2 = NULL;
+			SDL_PushEvent(&event);
+
+			bus_watch_id = 0;
+			return FALSE;
+		}
+		break;
+		default:
+		break;
 	}
 
-	return interval;
+	return TRUE;
 }
 
 void CIN_Play()
@@ -141,6 +133,7 @@ void CIN_Play()
 
 	if (!pipeline) {
 		GstElement *playbin, *videosink;
+		GstBus *bus;
 
 		pipeline = gst_pipeline_new("cinema");
 		playbin = gst_element_factory_make("playbin2", "playbin");
@@ -152,7 +145,8 @@ void CIN_Play()
 			NULL);
 		gst_bin_add(GST_BIN(pipeline), playbin);
 
-		timer_id = SDL_AddTimer(TIMER_INTERVAL, timer_cb, NULL);
+		bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline));
+		bus_watch_id = gst_bus_add_watch(bus, bus_cb, NULL);
 	} else {
 		/* Resuming existing pipeline */
 	}
@@ -163,23 +157,12 @@ void CIN_Play()
 void CIN_Pause()
 {
 	TRACE("Pause");
-
 	gst_element_set_state(pipeline, GST_STATE_PAUSED);
 }
 
 void CIN_Stop()
 {
 	TRACE("Stop playing now");
-
-	if (pipeline) {
-		gst_element_set_state(pipeline, GST_STATE_NULL);
-		gst_object_unref(pipeline);
-		pipeline = NULL;
-	}
-
-	if (timer_id) {
-		SDL_RemoveTimer(timer_id);
-		timer_id = 0;
-	}
+	free_pipeline();
 }
 
