@@ -8,11 +8,12 @@
 #include "override.h"
 #include "debug.h"
 
-#define DEBUG_DOMAIN "ACCEL"
-#define ACCEL_FILE "/sys/class/i2c-adapter/i2c-3/3-001d/coord"
-#define ACCEL_BUF_SIZE 256
+#define DEBUG_DOMAIN    "ACCEL"
+#define CAL_FILE        "/etc/preenv/cal"
+#define ACCEL_FILE      "/sys/class/i2c-adapter/i2c-3/3-001d/coord"
+#define ACCEL_BUF_SIZE  256
 
-#define EVENT_INTERVAL 100
+#define EVENT_INTERVAL  100
 
 static struct _SDL_Joystick
 {
@@ -24,14 +25,53 @@ static struct _SDL_Joystick
 static SDL_bool poll = SDL_TRUE;
 static SDL_TimerID timer = 0;
 
-static inline Sint16 scale(long v)
+#define SINT16_MAX 32767
+#define SINT16_MIN -32768
+
+static int cal_offset[3] = { 0, 0, 0 };
+static int cal_divider[3] = { 1000, 1000, 1000 };
+
+static void read_calibration()
 {
-	return (v * 32767) / 1020;
+	static char buffer[ACCEL_BUF_SIZE+1];
+	int n, fd = open(CAL_FILE, O_RDONLY);
+	if (fd < 0) {
+		TRACE("Using default calibration values");
+		return;
+	}
+
+	if ((n = read(fd, &buffer, ACCEL_BUF_SIZE)) <= 0) {
+		WARN("Failed to read calibration file");
+		close(fd);
+		return;
+	}
+	buffer[n] = '\0';
+
+	if (sscanf(buffer, "%i %i %i %i %i %i",
+		&cal_offset[0], &cal_offset[1], &cal_offset[2],
+		&cal_divider[0], &cal_divider[1], &cal_divider[2]) == 6) {
+		TRACE("Using calibration values: %i %i %i %i %i %i",
+			cal_offset[0], cal_offset[1], cal_offset[2],
+			cal_divider[0], cal_divider[1], cal_divider[2]);
+	} else {
+		WARN("Failed to parse calibration file");
+	}
+
+	close(fd);
 }
 
-static inline void maybe_update(Uint8 axis, Sint16 * value, Sint16 new)
+static inline Sint16 scale(Uint8 axis, long v)
 {
-	Sint16 scaled = scale(new);
+	long s = ((v + cal_offset[axis]) * SINT16_MAX) / cal_divider[axis];
+	/* Clamp value */
+	if (s > SINT16_MAX) return SINT16_MAX;
+	if (s < SINT16_MIN) return SINT16_MIN;
+	return s;
+}
+
+static inline void maybe_update(Uint8 axis, Sint16 * value, int new)
+{
+	Sint16 scaled = scale(axis, new);
 	if (*value != scaled) {
 		if (poll) {
 			SDL_Event fakeevent;
@@ -122,6 +162,10 @@ const char *SDL_JoystickName(int index)
 SDL_Joystick *SDL_JoystickOpen(int index)
 {
 	if (index == 0) {
+		if (accel.opened) {
+			WARN("Accelerometer is already opened");
+			return &accel;
+		}
 		TRACE("Opening accelerometer");
 		accel.fd = open(ACCEL_FILE, O_RDONLY|O_NONBLOCK);
 		if (accel.fd < 0) {
@@ -129,6 +173,7 @@ SDL_Joystick *SDL_JoystickOpen(int index)
 			return NULL;
 		}
 		accel.opened = 1;
+		read_calibration();
 		update_polling();
 		return &accel;
 	} else {
