@@ -13,7 +13,7 @@
 		dlerror(); \
 		super_func = dlsym(RTLD_NEXT, STRINGIFY(func)); \
 		if ((dl_error = dlerror()) != NULL) { \
-			fprintf(stderr, "linker: %s\n", dl_error); \
+			fprintf(stderr, "LINKER: %s\n", dl_error); \
 			exit(EXIT_FAILURE); \
 		} \
 	}
